#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for conveyor.conveyor module
"""

# pylint: disable=R0201, W0212
import argparse
import copy
import logging
import sys
from http import HTTPStatus
from unittest.mock import MagicMock, patch

import conveyor.conveyor as con
from conveyor.retrieve import RetrievalConveyor
from pyramid.response import Response
from urllib3.connection import HTTPConnection

logger = logging.getLogger("test_conveyer")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

expected_settings = {
    "qa_delivery_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/qa2",
    "weblog_cache_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/cache/weblog",
    "workspaces_lustre_root_dir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool",
    "current_root_directory": "current/directory",
    "current_subdirectory": "dir_123",
    "destination_dir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpabcd1234",
    "destination_subdir": "tmpabcd1234",
    "system_id": 2,
    "sdm_ids": "brain_000.58099.67095825232",
    "workflow_url": "http://fake:1234",
}

args = argparse.Namespace(deliver=None, retrieve=None, deliver_img=None, retrieve_img=None, qa_fail=None)

TEST_JSON = "test/test.json"
PARSE_ARGS_PATCH = "argparse.ArgumentParser.parse_args"
CONVEY_PATCH = "conveyor.retrieve.RetrievalConveyor.convey"


class TestConveyor:
    """Tests for conveyor"""

    def test_get_settings(self):
        """
        Be sure all the settings needed get set.

        :return:
        """
        with patch("pathlib.Path.cwd") as cwd:
            settings = con._get_settings(TEST_JSON)
            assert settings["qa_delivery_area"] == expected_settings["qa_delivery_area"]
            assert settings["weblog_cache_area"] == expected_settings["weblog_cache_area"]
            assert settings["workspaces_lustre_root_dir"] == expected_settings["workspaces_lustre_root_dir"]
            # mock calls to cwd and count
            assert cwd.call_count == 2
            assert settings["destination_dir"] == expected_settings["destination_dir"]
            assert settings["destination_subdir"] == expected_settings["destination_subdir"]
            assert settings["request_id"] == expected_settings["system_id"]
            assert settings["sdm_ids"] == expected_settings["sdm_ids"]
            assert settings["workflow_url"] is not None

    def test_main_deliver(self):
        """
        Confirm that DeliveryConveyor actually conveys.

        :return:
        """
        args.deliver = [TEST_JSON]

        with patch(PARSE_ARGS_PATCH, MagicMock(return_value=args)):
            assert args.deliver[0] == TEST_JSON
            with patch("conveyor.deliver.DeliveryConveyor.convey") as del_convey:
                con.main()
                assert del_convey.call_count == 1

        # reset for other testing
        args.deliver = None

    def test_main_retrieve(self):
        """
        Confirm that DeliveryConveyor retrieves what it's meant to convey.

        :return:
        """
        args.retrieve = [TEST_JSON]

        with patch(PARSE_ARGS_PATCH, MagicMock(return_value=args)):
            with patch(CONVEY_PATCH) as ret_convey:
                con.main()
                assert ret_convey.call_count == 1

        # reset for other testing
        args.retrieve = None
