#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
.. codeauthor:: Nathan Hertz <nhertz@nrao.edu>

Workspaces testing utilities

Put things here if you'd like them to be available to any given test in the entire system

HOW TO USE:
1. In any given pytest-powered test, add this line below the imports:
pytest_plugins = ["testing.utils.conftest"]
2. All fixtures within this file will now be available for use
3. If you want a non-fixture item from here, you'll have to do a normal Python import
"""

# pylint: disable=C0103, E0401, R0903, W0621

import json
from datetime import datetime
from typing import Any, List, Union
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
from messaging.messenger import SenderIF
from messaging.router import Router
from mock_alchemy.mocking import UnifiedAlchemyMagicMock
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from workspaces.capability.schema import (
    Capability,
    CapabilityExecution,
    CapabilityRequest,
)
from workspaces.capability.schema import mapper_registry as capability_mapper_registry
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.capability_service import CapabilityService
from workspaces.capability.services.execution_manager import ExecutionManager
from workspaces.capability.services.state_machine_info import StateMachineInfo
from workspaces.capability.services.state_machine_info import (
    mapper_registry as state_machine_mapper_registry,
)
from workspaces.notification.schema import NotificationTemplate
from workspaces.notification.services.notification_info import NotificationInfo
from workspaces.notification.services.notification_service import NotificationService
from workspaces.workflow.schema import Workflow, WorkflowRequest
from workspaces.workflow.services.interfaces import WorkflowInfoIF
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import WorkflowService


@pytest.fixture(scope="module")
def mock_capability_info() -> CapabilityInfo:
    """
    Pytest fixture providing a mock CapabilityInfo object with a test capability; mocked object can be used
    like a normal CapabilityInfo instance

    :return: Mocked CapabilityInfo
    """

    engine = create_engine("sqlite:///:memory:")
    capability_mapper_registry.metadata.create_all(engine)
    session = Session(engine)

    mocked_capability_info = CapabilityInfo(session)

    yield mocked_capability_info


@pytest.fixture(scope="module")
def mock_state_machine_info() -> StateMachineInfo:
    """
    Pytest fixture providing a mock CapabilityInfo object with a test capability; mocked object can be used
    like a normal CapabilityInfo instance

    :return: Mocked CapabilityInfo
    """

    engine = create_engine("sqlite:///:memory:")
    capability_mapper_registry.metadata.create_all(engine)
    state_machine_mapper_registry.metadata.create_all(engine)
    session = Session(engine)

    mocked_state_machine_info = StateMachineInfo(session)
    mocked_state_machine_info.init_state_machines = MagicMock()

    yield mocked_state_machine_info


@pytest.fixture(scope="module")
def mock_capabilities(mock_capability_info: CapabilityInfo) -> List[Capability]:
    null_capability = Capability(
        name="null",
        max_jobs=2,
        enabled=True,
        has_image_products=True,
    )
    test_capability = Capability(name="test", max_jobs=777, enabled=False, has_image_products=False)

    add_data_to_db_mock(mock_capability_info, null_capability)
    add_data_to_db_mock(mock_capability_info, test_capability)

    return [null_capability, test_capability]


def add_data_to_db_mock(db_mock: Union[CapabilityInfo, WorkflowInfoIF], data: Any):
    """
    Add a query that returns data to a mocked database access object

    :param db_mock: Mocked data access object (either CapabilityInfo or WorkflowInfo)
    :param query: Query that should return the given data
        EX: session.query(Capability).filter_by(name=capability_name).first() turns into
            [mock.call.query(Capability), mock.call.filter_by(name=capability_name), mock.call.first()]
    :param data: List of data to return from query
    """
    db_mock.session.add(data)


@pytest.fixture(scope="module")
def mock_capability_service(
    mock_capability_info: CapabilityInfo,
    mock_state_machine_info: StateMachineInfo,
    mock_workflow_service: WorkflowService,
    mock_notification_service: NotificationService,
) -> CapabilityService:
    """
    Pytest fixture providing a mock CapabilityService object with no threaded event listener creation

    :return: Mocked CapabilityService
    """
    with patch("workspaces.capability.services.capability_service.Router"):
        with patch("workspaces.capability.services.capability_service.MessageSender"):
            return CapabilityService(
                mock_capability_info, mock_state_machine_info, mock_workflow_service, mock_notification_service
            )


@pytest.fixture(scope="module")
def mock_workflow_info() -> WorkflowInfo:
    """
    Pytest fixture providing a mock WorkflowInfo object; mocked object can be used
    like a normal WorkflowInfo instance

    :return: Mocked WorkflowInfo
    """

    null_workflow = Workflow(
        workflow_name="null",
    )

    null_workflow_query = (
        [mock.call.query(Workflow), mock.call.filter_by(name="null")],
        [null_workflow],
    )

    mock_session = UnifiedAlchemyMagicMock(data=[null_workflow_query])
    mock_workflow_info = WorkflowInfo(mock_session)

    with patch(
        "workspaces.workflow.services.workflow_info.WorkflowInfo.lookup_workflow_definition", return_value=null_workflow
    ):
        with patch("workspaces.workflow.services.workflow_info.WorkflowInfo.save_request"):
            yield mock_workflow_info


@pytest.fixture(scope="module")
def mock_workflow_service(mock_workflow_info: WorkflowInfo) -> WorkflowService:
    """
    Pytest fixture providing a mock WorkflowService object

    :return: Mocked WorkflowService
    """
    workflow_message_handler = MagicMock()
    workflow_message_handler.archive_router = MagicMock()
    workflow_message_handler.message_router = MagicMock()
    with patch("workspaces.workflow.services.workflow_service.Router"):
        return WorkflowService(mock_workflow_info, workflow_message_handler)


@pytest.fixture(scope="module")
def mock_notification_info() -> NotificationInfo:
    """
    Pytest fixture providing a mock NotificationInfo object

    :return: Mocked NotificationInfo
    """
    email_notification = NotificationTemplate(
        name="email",
        description="A generic email template, requiring subject and message.",
        content="Subject: {{subject}}\n\n{{message}}",
    )

    mock_session = UnifiedAlchemyMagicMock()
    mock_notification_info = NotificationInfo(mock_session)
    mock_session.add(email_notification)
    with patch("workspaces.notification.services.notification_info.NotificationInfo.save_entity"):
        return mock_notification_info


@pytest.fixture(scope="module")
def mock_notification_service(mock_notification_info: NotificationInfo) -> NotificationService:
    """
    Pytest fixture providing a mock NotificationService object

    :return: Mocked NotificationService
    """
    return NotificationService(mock_notification_info)


@pytest.fixture(scope="module")
def mock_capability_requests(
    mock_capability_service: CapabilityService,
    mock_capability_info: CapabilityInfo,
) -> List[CapabilityRequest]:
    """
    Pytest fixture adding two mock CapabilityRequest objects that can be accessed from the mock_capability_info

    :param mock_capability_service: Mocked CapabilityService
    :param mock_capability_info: Mocked CapabilityInfo
    :return: Mocked CapabilityRequest
    """
    requests = []

    r1 = mock_capability_info.create_capability_request("null", ["-g"])
    r1.id = -1
    r1.ingested = False
    r1.created_at = datetime(1996, 9, 10)
    r1.updated_at = datetime(1996, 9, 10)
    requests.append(r1)

    add_data_to_db_mock(mock_capability_info, r1)

    r2 = mock_capability_info.create_capability_request("test")
    r2.id = -2
    r2.ingested = False
    r2.created_at = datetime(1996, 9, 10)
    r2.updated_at = datetime(1996, 9, 10)
    requests.append(r2)

    add_data_to_db_mock(mock_capability_info, r2)

    return requests


@pytest.fixture(scope="module")
def mock_workflow_requests(
    mock_workflow_service: WorkflowService, mock_workflow_info: WorkflowInfoIF
) -> List[WorkflowRequest]:
    """
    Make us some mock workflow requests for testing.

    :param mock_workflow_service:
    :param mock_workflow_info:
    :return:
    """
    requests = []

    r1 = WorkflowRequest(
        workflow_request_id=-1,
        workflow_name="null",
        argument={"parameters": "-g"},
        state="Ready",
        results_dir="",
        created_at=datetime(1996, 9, 10),
        updated_at=datetime(1996, 9, 10),
    )
    requests.append(r1)

    add_data_to_db_mock(mock_workflow_info, r1)

    r2 = WorkflowRequest(
        workflow_request_id=-2,
        workflow_name="std_calibration",
        argument={
            "product_locator": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",
            "user_email": "chausman@nrao.edu",
        },
        state="Created",
        results_dir="",
        created_at=datetime(2021, 1, 1),
        updated_at=datetime(2021, 1, 1),
    )
    requests.append(r2)

    add_data_to_db_mock(mock_workflow_info, r2)

    return requests


class MockMessageSender(SenderIF):
    """Pretends to send a message"""

    destination = None

    def send_message(self, **message):
        """
        Don't send this message, but play it on TV

        :param message: message to send as Dict
        :return:
        """
        print(f"sending message: {message}")
        self.destination.receive_message(json.dumps(message))


@pytest.fixture(scope="module")
def test_router() -> Router:
    """
    Pytest fixture providing a Router class for testing

    :return: Test Router
    """
    messenger = MockMessageSender()
    router = Router("test", messenger)
    messenger.destination = router
    return router


@pytest.fixture(scope="module")
def test_message_sender() -> SenderIF:
    """
    Pytest fixture providing a Router class for testing

    :return: Test Router
    """
    messenger = MockMessageSender()
    router = Router("test", messenger)
    messenger.destination = router
    return messenger


@pytest.fixture(scope="module")
def dummy_execution_manager(
    mock_capability_info: CapabilityInfo,
    mock_workflow_service: WorkflowService,
    test_router: Router,
    test_message_sender: SenderIF,
    mock_notification_service,
) -> ExecutionManager:
    """
    Pytest fixture that instantiates a dummy ExecutionManager object for testing purposes

    :return: Dummy ExecutionManager
    """
    return ExecutionManager(
        mock_capability_info, mock_workflow_service, test_router, test_message_sender, mock_notification_service
    )
